/// An integer-like type that will store small integers up to `i128` inline. Larger integers are
/// represented as a slice to a sequence of base 2<sup>32</sup> digits represented as a `*mut u32`.
#[derive(Clone, PartialEq, Eq)]
pub struct SmallInt(pub(crate) SmallIntType);

/// An integer-like type that will store small integers up to `u128` inline. Larger integers are
/// represented as a slice to a sequence of base 2<sup>32</sup> digits represented as a `*mut u32`.
#[derive(Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct SmallUint(pub(crate) SmallUintType);

#[derive(Clone, PartialEq, Eq)]
pub enum SmallIntType {
    Inline(i128),
    Heap((*mut u32, isize)),
}

#[derive(Clone, PartialEq, Eq, PartialOrd, Ord)]
pub enum SmallUintType {
    Inline(u128),
    Heap((*mut u32, usize)),
}

impl Drop for SmallInt {
    fn drop(&mut self) {
        if let Self(SmallIntType::Heap((r, s))) = self {
            let size = usize::try_from(s.abs()).unwrap();
            let slice = unsafe { core::slice::from_raw_parts_mut(*r, size) };
            unsafe { std::mem::drop(Box::from_raw(slice)) }
        }
    }
}

impl Drop for SmallUint {
    fn drop(&mut self) {
        if let Self(SmallUintType::Heap((r, s))) = self {
            let slice = unsafe { core::slice::from_raw_parts_mut(*r, *s) };
            unsafe { std::mem::drop(Box::from_raw(slice)) }
        }
    }
}
